# encoding: utf-8

import os
import json
from collections import OrderedDict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import Formatter

from jaqs.trade.analyze.report import Report
from jaqs.data import RemoteDataService
from jaqs.data.basic.instrument import InstManager
import jaqs.util as jutil

STATIC_FOLDER = jutil.join_relative_path("trade/analyze/static")


class MyFormatter(Formatter):
    def __init__(self, dates, fmt='%Y%m'):
        self.dates = dates
        self.fmt = fmt

    def __call__(self, x, pos=0):
        """Return the label for time x at position pos"""
        ind = int(np.round(x))
        if ind >= len(self.dates) or ind < 0:
            return ''

        # return self.dates[ind].strftime(self.fmt)
        return pd.to_datetime(self.dates[ind], format="%Y%m%d").strftime(self.fmt)


class BaseAnalyzer(object):
    """
    Attributes
    ----------
    _trades : pd.DataFrame
    _configs : dict
    data_api : BaseDataServer
    _universe : set
        All securities that have been traded.
        
    """
    def __init__(self):
        self.file_folder = ""
        
        self._trades = None
        self._configs = None
        self.data_api = None
        self.dataview = None
        
        self._universe = []
        self._closes = None
        self.daily_position = None
        
        self.adjust_mode = None
        
        self.inst_map = dict()
        
        self.metrics = dict()
        
        self.report_dic = dict()
        
    @property
    def trades(self):
        """Read-only attribute"""
        return self._trades
    
    @property
    def universe(self):
        """Read-only attribute"""
        return self._universe
    
    @property
    def configs(self):
        """Read-only attribute"""
        return self._configs
    
    @property
    def closes(self):
        """Read-only attribute, close prices of securities in the universe"""
        return self._closes
    
    def initialize(self, data_api=None, dataview=None, file_folder='.'):
        """
        Read trades from csv file to DataFrame of given data type.

        Parameters
        ----------
        data_api : RemoteDataService
        dataview : DataView
        file_folder : str
            Directory path where trades and configs are stored.

        """
        self.data_api = data_api
        self.dataview = dataview
        
        type_map = {'task_id': str,
                    'entrust_no': str,
                    'entrust_action': str,
                    'symbol': str,
                    'fill_price': float,
                    'fill_size': float,
                    'fill_date': int,
                    'fill_time': int,
                    'fill_no': str,
                    'commission': float}
        abs_path = os.path.abspath(file_folder)
        self.file_folder = abs_path
        trades = pd.read_csv(os.path.join(self.file_folder, 'trades.csv'), ',', dtype=type_map)
        
        self._init_universe(trades.loc[:, 'symbol'].values)
        self._init_configs(self.file_folder)
        self._init_trades(trades)
        self._init_symbol_price()
        self._init_inst_data()
    
    def _init_inst_data(self):
        symbol_str = ','.join(self.universe)
        if self.dataview is not None:
            data_inst = self.dataview.data_inst
            self.inst_map = data_inst.to_dict(orient='index')
        elif self.data_api is not None:
            inst_mgr = InstManager(data_api=self.data_api, symbol=symbol_str)
            self.inst_map = {k: v.__dict__ for k, v in inst_mgr.inst_map.items()}
            del inst_mgr
        else:
            raise ValueError("no dataview or dataapi provided.")
        
    def _init_trades(self, df):
        """Add datetime column. """
        df.loc[:, 'fill_dt'] = jutil.combine_date_time(df.loc[:, 'fill_date'], df.loc[:, 'fill_time'])
        
        df = df.set_index(['symbol', 'fill_dt']).sort_index(axis=0)
        
        # self._trades = jutil.group_df_to_dict(df, by='symbol')
        self._trades = df
    
    def _init_symbol_price(self):
        """Get close price of securities in the universe from data server."""
        if self.dataview is not None:
            df_close = self.dataview.get_ts('close', start_date=self.start_date, end_date=self.end_date)
        else:
            df, msg = self.data_api.daily(symbol=','.join(self.universe), fields='trade_date,symbol,close',
                                          start_date=self.start_date, end_date=self.end_date)
            if msg != '0,':
                print(msg)
            df_close = df.pivot(index='trade_date', columns='symbol', values='close')
        self._closes = df_close

    def _init_universe(self, securities):
        """Return a set of securities."""
        self._universe = set(securities)
    
    def _init_configs(self, folder):
        import codecs
        with codecs.open(os.path.join(folder, 'configs.json'), 'r', encoding='utf-8') as f:
            configs = json.load(f)
        self._configs = configs
        self.init_balance = self.configs['init_balance']
        self.start_date = self.configs['start_date']
        self.end_date = self.configs['end_date']
        
    @staticmethod
    def _process_trades(df):
        """Add various statistics to trades DataFrame."""
        from jaqs.trade import common
        
        # df = df.set_index('fill_date')
    
        # pre-process
        cols_to_drop = ['task_id', 'entrust_no', 'fill_no']
        df = df.drop(cols_to_drop, axis=1)
        
        def _apply(gp_df):
            # calculation of non-cumulative fields
            direction = gp_df['entrust_action'].apply(lambda s: 1 if common.ORDER_ACTION.is_positive(s) else -1)
            fill_size, fill_price = gp_df['fill_size'], gp_df['fill_price']
            turnover = fill_size * fill_price

            gp_df.loc[:, 'BuyVolume'] = (direction + 1) / 2 * fill_size
            gp_df.loc[:, 'SellVolume'] = (direction - 1) / -2 * fill_size
            
            # Calculation of cumulative fields
            gp_df.loc[:, 'CumVolume'] = fill_size.cumsum()
            gp_df.loc[:, 'CumTurnOver'] = turnover.cumsum()
            gp_df.loc[:, 'CumNetTurnOver'] = (turnover * -direction).cumsum()
    
            gp_df.loc[:, 'position'] = (fill_size * direction).cumsum()
    
            gp_df.loc[:, 'AvgPosPrice'] = calc_avg_pos_price(gp_df.loc[:, 'position'].values, fill_price.values)
    
            gp_df.loc[:, 'CumProfit'] = (gp_df.loc[:, 'CumNetTurnOver'] + gp_df.loc[:, 'position'] * fill_price)
            return gp_df
        gp = df.groupby(by='symbol')
        res = gp.apply(_apply)
        
        return res
    
    def process_trades(self):
        # self._trades = {k: self._process_trades(v) for k, v in self.trades.viewitems()}
        self._trades = self._process_trades(self._trades)
    
    def get_pos_change_info(self):
        trades = pd.concat(self.trades.values(), axis=0)
        gp = trades.groupby(by=['fill_date'], as_index=False)
        res = OrderedDict()
        account = OrderedDict()
    
        for date, df in gp:
            df_mod = df.loc[:, ['symbol', 'entrust_action', 'fill_size', 'fill_price',
                                'position', 'AvgPosPrice']]
            df_mod.columns = ['symbol', 'action', 'size', 'price',
                              'position', 'cost price']
        
            res[str(date)] = df_mod
        
            mv = sum(df_mod.loc[:, 'price'] * df.loc[:, 'position'])
            current_profit = sum(df.loc[:, 'CumProfit'])
            cash = self.configs['init_balance'] + current_profit - mv
        
            account[str(date)] = {'market_value': mv, 'cash': cash}
        self.position_change = res
        self.account = account

    def get_daily(self):
        close = self.closes
        trade = self.trades
        
        # pro-process
        trade_cols = ['fill_date', 'BuyVolume', 'SellVolume', 'commission', 'position', 'AvgPosPrice', 'CumNetTurnOver']
    
        trade = trade.loc[:, trade_cols]
        gp = trade.groupby(by=['symbol', 'fill_date'])
        func_last = lambda ser: ser.iat[-1]
        trade = gp.agg({'BuyVolume': np.sum, 'SellVolume': np.sum, 'commission': np.sum,
                        'position': func_last, 'AvgPosPrice': func_last, 'CumNetTurnOver': func_last})
        trade.index.names = ['symbol', 'trade_date']

        # get daily position
        df_position = trade['position'].unstack('symbol').fillna(method='ffill').fillna(0.0)
        daily_position = df_position.reindex(close.index)
        daily_position = daily_position.fillna(method='ffill').fillna(0)
        self.daily_position = daily_position
        
        # calculate statistics
        close = pd.DataFrame(close.T.stack())
        close.columns = ['close']
        close.index.names = ['symbol', 'trade_date']
        merge = pd.concat([close, trade], axis=1, join='outer')
        
        def _apply(gp_df):
            cols_nan_to_zero = ['BuyVolume', 'SellVolume', 'commission']
            cols_nan_fill = ['close', 'position', 'AvgPosPrice', 'CumNetTurnOver']
            # merge: pd.DataFrame
            gp_df.loc[:, cols_nan_fill] = gp_df.loc[:, cols_nan_fill].fillna(method='ffill')
            gp_df.loc[:, cols_nan_fill] = gp_df.loc[:, cols_nan_fill].fillna(0)
    
            gp_df.loc[:, cols_nan_to_zero] = gp_df.loc[:, cols_nan_to_zero].fillna(0)
    
            mask = gp_df.loc[:, 'AvgPosPrice'] < 1e-5
            gp_df.loc[mask, 'AvgPosPrice'] = gp_df.loc[mask, 'close']
    
            gp_df.loc[:, 'CumProfit'] = gp_df.loc[:, 'CumNetTurnOver'] + gp_df.loc[:, 'position'] * gp_df.loc[:, 'close']
            gp_df.loc[:, 'CumProfitComm'] = gp_df['CumProfit'] - gp_df['commission'].cumsum()
    
            daily_net_turnover = gp_df['CumNetTurnOver'].diff(1).fillna(gp_df['CumNetTurnOver'].iat[0])
            daily_position_change = gp_df['position'].diff(1).fillna(gp_df['position'].iat[0])
            gp_df['trading_pnl'] = (daily_net_turnover + gp_df['close'] * daily_position_change)
            gp_df['holding_pnl'] = (gp_df['close'].diff(1) * gp_df['position'].shift(1)).fillna(0.0)
            gp_df.loc[:, 'total_pnl'] = gp_df['trading_pnl'] + gp_df['holding_pnl']
            
            return gp_df

        gp = merge.groupby(by='symbol')
        res = gp.apply(_apply)
        
        self.daily = res

    '''
    def get_daily(self):
        """Add various statistics to daily DataFrame."""
        self.daily = self._get_daily(self.closes, self.trades)
        daily_dic = dict()
        for sec, df_trade in self.trades.viewitems():
            df_close = self.closes[sec].rename('close')
        
            res = self._get_daily(df_close, df_trade)
            daily_dic[sec] = res
    
        self.daily = daily_dic
    '''

    def get_returns(self, compound_return=True, consider_commission=True):
        cols = ['trading_pnl', 'holding_pnl', 'total_pnl', 'commission', 'CumProfitComm', 'CumProfit']
        '''
        dic_symbol = {sec: self.inst_map[sec]['multiplier'] * df_daily.loc[:, cols]
                      for sec, df_daily in self.daily.items()}
                
        df_profit = pd.concat(dic_symbol, axis=1)  # this is cumulative profit
        df_profit = df_profit.fillna(method='ffill').fillna(0.0)
        df_pnl = df_profit.stack(level=1)
        df_pnl = df_pnl.sum(axis=1)
        df_pnl = df_pnl.unstack(level=1)
        '''
        
        daily = self.daily.loc[:, cols]
        daily = daily.stack().unstack('symbol')
        
        df_pnl = daily.sum(axis=1)
        df_pnl = df_pnl.unstack(level=1)

        self.df_pnl = df_pnl
    
        # TODO temperary solution
        if consider_commission:
            strategy_value = (df_pnl['total_pnl'] - df_pnl['commission']).cumsum() + self.init_balance
        else:
            strategy_value = df_pnl['total_pnl'].cumsum() + self.init_balance
    
        market_values = pd.concat([strategy_value, self.data_benchmark], axis=1).fillna(method='ffill')
        market_values.columns = ['strat', 'bench']
    
        df_returns = market_values.pct_change(periods=1).fillna(0.0)
    
        df_returns = df_returns.join((df_returns.loc[:, ['strat', 'bench']] + 1.0).cumprod(), rsuffix='_cum')
        if compound_return:
            df_returns.loc[:, 'active_cum'] = df_returns['strat_cum'] - df_returns['bench_cum'] + 1
            df_returns.loc[:, 'active'] = df_returns['active_cum'].pct_change(1).fillna(0.0)
        else:
            df_returns.loc[:, 'active'] = df_returns['strat'] - df_returns['bench']
            df_returns.loc[:, 'active_cum'] = df_returns['active'].add(1.0).cumprod(axis=0)
    
        start = pd.to_datetime(self.configs['start_date'], format="%Y%m%d")
        end = pd.to_datetime(self.configs['end_date'], format="%Y%m%d")
        years = (end - start).days / 365.0
    
        self.metrics['yearly_return'] = np.power(df_returns.loc[:, 'active_cum'].values[-1], 1. / years) - 1
        self.metrics['yearly_vol'] = df_returns.loc[:, 'active'].std() * np.sqrt(225.)
        self.metrics['beta'] = np.corrcoef(df_returns.loc[:, 'bench'], df_returns.loc[:, 'strat'])[0, 1]
        self.metrics['sharpe'] = self.metrics['yearly_return'] / self.metrics['yearly_vol']
    
        # bt_strat_mv = pd.read_csv('bt_strat_mv.csv').set_index('trade_date')
        # df_returns = df_returns.join(bt_strat_mv, how='right')
        self.returns = df_returns
    
    def plot_pnl(self, save_folder=None):
        if save_folder is None:
            save_folder = self.file_folder
        
        fig, (ax0, ax1, ax2) = plt.subplots(3, 1, figsize=(21, 8), dpi=300, sharex=True)
        idx0 = self.returns.index
        idx = range(len(idx0))
        
        ax0.plot(idx, self.df_pnl['trading_pnl'], lw=1.5, color='indianred', label='Trading PnL')
        ax0.plot(idx, self.df_pnl['holding_pnl'], lw=1.5, color='royalblue', label='Holding PnL')
        ax0.axhline(0.0, color='k', lw=1, ls='--')
        # ax0.plot(idx, self.pnl['total_pnl'], lw=1.5, color='violet', label='Total PnL')
        ax0.legend(loc='upper left')
        
        ax1.plot(idx, self.returns.loc[:, 'bench_cum'], label='Benchmark')
        ax1.plot(idx, self.returns.loc[:, 'strat_cum'], label='Strategy')
        ax1.legend(loc='upper left')
        
        ax2.plot(idx, self.returns.loc[:, 'active_cum'], label='Extra Return')
        ax2.legend(loc='upper left')
        ax2.set_xlabel("Date")
        ax2.set_ylabel("Percent")
        ax1.set_ylabel("Percent")
        ax2.xaxis.set_major_formatter(MyFormatter(idx0, '%Y-%m-%d'))
    
        plt.tight_layout()
        fig.savefig(os.path.join(save_folder, 'pnl_img.png'))
        plt.close()

    def gen_report(self, source_dir, template_fn, out_folder='.', selected=None):
        """
        Generate HTML (and PDF) report of the trade analysis.

        Parameters
        ----------
        source_dir : str
            path of directory where HTML template and css files are stored.
        template_fn : str
            File name of HTML template.
        out_folder : str
            Output folder of report.
        selected : list of str or None
            List of symbols whose detailed PnL curve and position will be plotted.
            # TODO: this parameter should not belong to function


        """
        dic = dict()
        dic['html_title'] = "Alpha Strategy Backtest Result"
        dic['selected_securities'] = selected
        dic['props'] = self.configs
        dic['metrics'] = self.metrics
        dic['position_change'] = self.position_change
        dic['account'] = self.account
        dic['df_daily'] = jutil.group_df_to_dict(self.daily, by='symbol')
        
        self.report_dic.update(dic)
        
        self.returns.to_csv(os.path.join(out_folder, 'returns.csv'))
    
        r = Report(self.report_dic, source_dir=source_dir, template_fn=template_fn, out_folder=out_folder)
        
        r.generate_html()
        r.output_html('report.html')

    def do_analyze(self, result_dir, selected_sec=None):
        if selected_sec is None:
            selected_sec = []
            
        print "process trades..."
        self.process_trades()
        print "get daily stats..."
        self.get_daily()
        print "calc strategy return..."
        self.get_returns(consider_commission=False)

        if len(selected_sec) > 0:
            print "Plot single securities PnL"
            for symbol in selected_sec:
                df_daily = self.daily.get(symbol, None)
                if df_daily is not None:
                    plot_trades(df_daily, symbol=symbol)

        print "Plot strategy PnL..."
        self.plot_pnl(result_dir)
        
        print "generate report..."
        self.gen_report(source_dir=STATIC_FOLDER, template_fn='report_template.html',
                        out_folder=result_dir,
                        selected=selected_sec)
        


class EventAnalyzer(BaseAnalyzer):
    def __init__(self):
        super(EventAnalyzer, self).__init__()
        
        self.metrics = dict()
        self.daily = None
        self.data_benchmark = None
        
        self.returns = None  # OrderedDict
        self.position_change = None  # OrderedDict
        self.account = None  # OrderedDict
        
    def initialize(self, data_server_=None, dataview=None, file_folder='.'):
        super(EventAnalyzer, self).initialize(data_api=data_server_, dataview=dataview,
                                              file_folder=file_folder)
        if self.dataview is not None and self.dataview.data_benchmark is not None:
            self.data_benchmark = self.dataview.data_benchmark.loc[(self.dataview.data_benchmark.index >= self.start_date)
                                                                   &(self.dataview.data_benchmark.index <= self.end_date)]
        else:
            self.data_benchmark = pd.DataFrame(index=self.closes.index, columns=['bench'], data=np.ones(len(self.closes), dtype=float))
            

class AlphaAnalyzer(BaseAnalyzer):
    def __init__(self):
        super(AlphaAnalyzer, self).__init__()
        
        self.metrics = dict()
        self.daily = None
        self.returns = None  # OrderedDict
        self.position_change = None  # OrderedDict
        self.account = None  # OrderedDict
        
        self.data_benchmark = None

    def initialize(self, data_api=None, dataview=None, file_folder='.'):
        super(AlphaAnalyzer, self).initialize(data_api=data_api, dataview=dataview,
                                              file_folder=file_folder)
        if self.dataview is not None and self.dataview.data_benchmark is not None:
            self.data_benchmark = self.dataview.data_benchmark.loc[(self.dataview.data_benchmark.index >= self.start_date)
                                                                   &(self.dataview.data_benchmark.index <= self.end_date)]
    
    @staticmethod
    def _to_pct_return(arr, cumulative=False):
        """Convert portfolio value to portfolio (linear) return."""
        r = np.empty_like(arr)
        r[0] = 0.0
        if cumulative:
            r[1:] = arr[1:] / arr[0] - 1
        else:
            r[1:] = arr[1:] / arr[:-1] - 1
        return r

    '''
    def get_returns_OLD(self, compound_return=True, consider_commission=True):
        profit_col_name = 'CumProfitComm' if consider_commission else 'CumProfit'
        vp_list = {sec: df_profit.loc[:, profit_col_name] for sec, df_profit in self.daily.viewitems()}
        df_profit = pd.concat(vp_list, axis=1)  # this is cumulative profit
        # TODO temperary solution
        df_profit = df_profit.fillna(method='ffill').fillna(0.0)
        strategy_value = df_profit.sum(axis=1) + self.configs['init_balance']
        
        market_values = pd.concat([strategy_value, self.data_benchmark], axis=1).fillna(method='ffill')
        market_values.columns = ['strat', 'bench']
        
        df_returns = market_values.pct_change(periods=1).fillna(0.0)
        
        df_returns = df_returns.join((df_returns.loc[:, ['strat', 'bench']] + 1.0).cumprod(), rsuffix='_cum')
        if compound_return:
            df_returns.loc[:, 'active_cum'] = df_returns['strat_cum'] - df_returns['bench_cum'] + 1
            df_returns.loc[:, 'active'] = df_returns['active_cum'].pct_change(1).fillna(0.0)
        else:
            df_returns.loc[:, 'active'] = df_returns['strat'] - df_returns['bench']
            df_returns.loc[:, 'active_cum'] = df_returns['active'].add(1.0).cumprod(axis=0)
        
        start = pd.to_datetime(self.configs['start_date'], format="%Y%m%d")
        end = pd.to_datetime(self.configs['end_date'], format="%Y%m%d")
        years = (end - start).days / 365.0
        
        self.metrics['yearly_return'] = np.power(df_returns.loc[:, 'active_cum'].values[-1], 1. / years) - 1
        self.metrics['yearly_vol'] = df_returns.loc[:, 'active'].std() * np.sqrt(225.)
        self.metrics['beta'] = np.corrcoef(df_returns.loc[:, 'bench'], df_returns.loc[:, 'strat'])[0, 1]
        self.metrics['sharpe'] = self.metrics['yearly_return'] / self.metrics['yearly_vol']
        
        # bt_strat_mv = pd.read_csv('bt_strat_mv.csv').set_index('trade_date')
        # df_returns = df_returns.join(bt_strat_mv, how='right')
        self.returns = df_returns

    '''
    
    def _get_index_weight(self):
        if self.dataview is not None:
            res = self.dataview.get_ts('index_weight', start_date=self.start_date, end_date=self.end_date)
        else:
            res = self.data_api.get_index_weights_daily(self.universe, self.start_date, self.end_date)
        return res
    
    def _brinson(self, close, pos, index_weight, group):
        """
        Brinson Attribution.
        
        Parameters
        ----------
        close : pd.DataFrame
            Index is date, columns are symbols.
        pos : pd.DataFrame
            Index is date, columns are symbols.
        index_weight : pd.DataFrame
            Index is date, columns are symbols.
        group : pd.DataFrame
            Index is date, columns are symbols.

        Returns
        -------
        dict

        """
        def group_sum(df, group_daily):
            groups = np.unique(group_daily.values.flatten())
            mask = np.isnan(groups.astype(float))
            groups = groups[np.logical_not(mask)]
            res = pd.DataFrame(index=df.index, columns=groups, data=np.nan)
            for g in groups:
                mask = group_daily == g
                tmp = df[mask]
                res.loc[:, g] = tmp.sum(axis=1)
            return res

        ret = close.pct_change(1)

        pf_weight = pos.div(pos.sum(axis=1), axis=0)

        weighted_ret_pf = ret.mul(pf_weight)
        weighted_ret_index = ret.mul(index_weight)

        index_group_weight = group_sum(index_weight, group)
        pf_group_weight = group_sum(pf_weight, group)

        pf_group_ret = group_sum(weighted_ret_pf, group).div(pf_group_weight)
        index_group_ret = group_sum(weighted_ret_index, group).div(index_group_weight)

        allo_ret_group = (pf_group_weight - index_group_weight).mul(index_group_ret)
        allo_ret = allo_ret_group.sum(axis=1)

        selection_ret_group = (pf_group_ret - index_group_ret).mul(index_group_weight)
        selection_ret = selection_ret_group.sum(axis=1)

        active_ret = (weighted_ret_pf.sum(axis=1) - weighted_ret_index.sum(axis=1))
        inter_ret = active_ret - selection_ret - allo_ret
    
        df_brinson = pd.DataFrame(index=allo_ret.index,
                                  data={'allocation': allo_ret,
                                        'selection': selection_ret,
                                        'interaction': inter_ret,
                                        'total_active': active_ret})
        
        return {'df_brinson': df_brinson, 'allocation': allo_ret_group, 'selection': selection_ret_group}
    
    def brinson(self, group):
        """
        
        Parameters
        ----------
        group : str or pd.DataFrame
            If group is string, this function will try to fetch the corresponding DataFrame from DataView.
            If group is pd.DataFrame, it will be used as-is.

        Returns
        -------

        """
        if isinstance(group, (str, unicode)):
            group = self.dataview.get_ts(group, start_date=self.start_date, end_date=self.end_date)
        elif isinstance(group, pd.DataFrame):
            pass
        else:
            raise ValueError("Group must be string or DataFrame. But {} is provided.".format(group))
        
        if group is None or group.empty:
            raise ValueError("group is None or group is empty")
        
        close = self.closes
        pos = self.daily_position
        index_weight = self._get_index_weight()
        
        res_dic = self._brinson(close, pos, index_weight, group)
        
        df_brinson = res_dic['df_brinson']
        self.report_dic['df_brinson'] = df_brinson
        plot_brinson(df_brinson, save_folder=self.file_folder)

    def do_analyze(self, result_dir, selected_sec=None, brinson_group=None):
        if selected_sec is None:
            selected_sec = []
    
        print "process trades..."
        self.process_trades()
        print "get daily stats..."
        self.get_daily()
        print "calc strategy return..."
        self.get_returns(consider_commission=False)
    
        not_none_sec = []
        if len(selected_sec) > 0:
            print "Plot single securities PnL"
            for symbol in selected_sec:
                df_daily = self.daily.loc[pd.IndexSlice[symbol, :], :]
                df_daily.index = df_daily.index.droplevel(0)
                if df_daily is not None:
                    not_none_sec.append(symbol)
                    plot_trades(df_daily, symbol=symbol, save_folder=self.file_folder)
    
        print "Plot strategy PnL..."
        self.plot_pnl(result_dir)
        
        if brinson_group is not None:
            print("Do brinson attribution.")
            group = self.dataview.get_ts(brinson_group)
            if group is None:
                raise ValueError("group data is None.")
            self.brinson(group)
    
        print "generate report..."
        self.gen_report(source_dir=STATIC_FOLDER, template_fn='report_template.html',
                        out_folder=result_dir,
                        selected=not_none_sec)


def plot_brinson(df, save_folder):
    """
    
    Parameters
    ----------
    df : pd.DataFrame

    """
    allo, selec, inter, total = df['allocation'], df['selection'], df['interaction'], df['total_active']
    fig, ax1 = plt.subplots(1, 1, figsize=(21, 8), dpi=300)
    
    idx0 = df.index
    idx = range(len(idx0))

    ax1.plot(idx, selec, lw=1.5, color='indianred', label='Selection Return')
    ax1.plot(idx, allo, lw=1.5, color='royalblue', label='Allocation Return')
    ax1.plot(idx, inter, lw=1.5, color='purple', label='Interaction Return')
    # ax1.plot(idx, total, lw=1.5, ls='--', color='k', label='Total Active Return')
    
    ax1.axhline(0.0, color='k', lw=0.5, ls='--')
    
    ax1.legend(loc='upper left')
    ax1.set_xlabel("Date")
    ax1.set_ylabel("Return")
    ax1.xaxis.set_major_formatter(MyFormatter(idx0, '%Y-%m-%d'))

    plt.tight_layout()
    fig.savefig(os.path.join(save_folder, 'brinson_attribution.png'))
    plt.close()
    



def calc_avg_pos_price(pos_arr, price_arr):
    """
    Calculate average cost price using position and fill price.
    When position = 0, cost price = symbol price.
    """
    assert len(pos_arr) == len(price_arr)

    avg_price = np.zeros_like(pos_arr, dtype=float)
    avg_price[0] = price_arr[0]
    for i in range(pos_arr.shape[0] - 1):
        if pos_arr[i+1] == 0:
            avg_price[i+1] = 0.0
        else:
            pos_diff = pos_arr[i+1] - pos_arr[i]
            if pos_arr[i] == 0 or pos_diff * pos_arr[i] > 0:
                count = True
            else:
                count = False
        
            if count:
                avg_price[i+1] = (avg_price[i] * pos_arr[i] + pos_diff * price_arr[i+1]) * 1. / pos_arr[i+1]
            else:
                avg_price[i+1] = avg_price[i]
    return avg_price


def plot_trades(df, symbol="", save_folder='.'):
    idx0 = df.index
    idx = range(len(idx0))
    price = df.loc[:, 'close']
    bv, sv = df.loc[:, 'BuyVolume'].values, df.loc[:, 'SellVolume'].values
    profit = df.loc[:, 'CumProfit'].values
    avgpx = df.loc[:, 'AvgPosPrice']
    bv *= .05
    sv *= .05
    
    fig = plt.figure(figsize=(14, 10), dpi=300)
    ax1 = plt.subplot2grid((4, 1), (0, 0), rowspan=3)
    ax3 = plt.subplot2grid((4, 1), (3, 0), rowspan=1, sharex=ax1)
    
    # fig, (ax1, ax3) = plt.subplots(2, 1, figsize=(16, 18), sharex=True)
    # fig, ax1 = plt.subplots(1, 1, figsize=(16, 6))
    ax2 = ax1.twinx()
    
    ax1.plot(idx, price, label='Price', linestyle='-', lw=1, marker='', color='yellow')
    ax1.scatter(idx, price, label='buy', marker='o', s=bv, color='red')
    ax1.scatter(idx, price, label='sell', marker='o', s=sv, color='green')
    ax1.plot(idx, avgpx, lw=1, marker='', color='green')
    ax1.legend(loc='upper left')
    
    ax2.plot(idx, profit, label='PnL', color='k', lw=1, ls='--', alpha=.4)
    ax2.legend(loc='upper right')
    
    # ax1.xaxis.set_major_formatter(MyFormatter(df.index))#, '%H:%M'))
    
    ax3.plot(idx, df.loc[:, 'position'], marker='D', markersize=3, lw=2)
    ax3.axhline(0, color='k', lw=1)
    
    ax1.set_title(symbol)
    
    ax1.xaxis.set_major_formatter(MyFormatter(idx0, '%Y-%m'))
    
    fig.savefig(save_folder + '/' + "{}.png".format(symbol))
    plt.tight_layout()
    return




# encoding: UTF-8
"""
Our trade is powered by this event engine that register, generate and process
events in an efficient way.
"""
from .engine import EventEngine, Event
from .eventtype import EVENT_TYPE



# encoding:utf-8

from .order import Order


class Trade(object):
    """
    Basic order class.

    Attributes
    ----------
    task_id : str
        id of the task.
    entrust_no : str
        ID of the order.
    entrust_action : str
    symbol : str
    fill_price : float
    fill_size : int
    fill_date : int
    fill_time : int
    fill_no : str
        ID of this trade.

    Methods
    -------


    """
    
    def __init__(self, order=None):
        self.task_id = 0
        self.entrust_no = ""
        
        self.entrust_action = ""
        
        self.symbol = ""

        self.fill_no = ""
        self.fill_price = 0.0
        self.fill_size = 0
        self.fill_date = 0
        self.fill_time = 0
        
        self.commission = 0.0

        if order is not None:
            if isinstance(order, Order):
                self.init_from_order(order)
            else:
                raise ValueError("init_from_order only accept argument of type Order.")
        
    def init_from_order(self, order):
        self.task_id = order.task_id
        self.entrust_no = order.entrust_no
        self.symbol = order.symbol
        self.entrust_action = order.entrust_action
    
    def set_fill_info(self, price, size, date, time, no):
        self.fill_price = price
        self.fill_size = size
        self.fill_date = date
        self.fill_time = time
        self.fill_no = no
        
    @classmethod
    def create_from_dict(cls, dic):
        trade_ind = cls()
        trade_ind.__dict__.update(dic)
        return trade_ind
    
    def __repr__(self):
        return "{0.fill_date:8d}({0.fill_time:8d}) " \
               "{0.entrust_action:6s} {0.symbol:10s}@{0.fill_price:.3f}" \
               "  size = {0.fill_size}".format(self)
    
    def __str__(self):
        return self.__repr__()


class TaskInd(object):
    def __init__(self, task_id=0, task_status="", task_algo="", task_msg=""):
        self.task_id = task_id
        self.task_status = task_status
        self.task_algo = task_algo
        self.task_msg = task_msg

    @classmethod
    def create_from_dict(cls, dic):
        rsp = cls()
        rsp.__dict__.update(dic)
        return rsp
    
    def __repr__(self):
        return "task_id = {0.task_id:d}  |  task_status = {0.task_status:s}  |  task_algo = {0.task_algo:s}" \
               "\ntask_msg = \n{0.task_msg:s}".format(self)

    def __str__(self):
        return self.__repr__()


class TradeStat(object):
    def __init__(self, symbol=""):
        self.symbol = symbol
        self.buy_filled_size = 0
        self.buy_want_size = 0
        self.sell_filled_size = 0
        self.sell_want_size = 0
    
    def __repr__(self):
        return ("        Want Size      Filled Size  \n"
                "====================================\n"
                "Buy     {0:8.0f}         {1:8.0f}   \n"
                "------------------------------------\n"
                "Sell    {2:8.0f}         {3:8.0f}   \n"
                "".format(self.buy_want_size, self.buy_filled_size,
                          self.sell_want_size, self.sell_filled_size))
    
    def __str__(self):
        return self.__repr__()



# encoding: utf-8
import pandas as pd
from jaqs.data import RemoteDataService
from jaqs.data import Parser
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)


def test_align():
    # -------------------------------------------------------------------------------------
    # input and pre-process demo data
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    raw, msg = ds.query_lb_fin_stat('income', '600000.SH', 20151225, 20170501, 'oper_rev')
    assert msg == '0,'
    
    idx_list = ['report_date', 'symbol']
    raw_idx = raw.set_index(idx_list)
    raw_idx.sort_index(axis=0, level=idx_list, inplace=True)
    
    df_ann = raw_idx.loc[pd.IndexSlice[:, :], 'ann_date']
    df_ann = df_ann.unstack(level=1)

    df_value = raw_idx.loc[pd.IndexSlice[:, :], 'oper_rev']
    df_value = df_value.unstack(level=1)
    
    date_arr = ds.get_trade_date_range(20160101, 20170501)
    df_close = pd.DataFrame(index=date_arr, columns=df_value.columns, data=1e3)
    
    # -------------------------------------------------------------------------------------
    # demo usage of parser
    parser = Parser()
    parser.register_function('Myfunc', lambda x: x * 0 + 1)  # simultaneously test register function and align
    expr_formula = 'signal / Myfunc(close)'
    expression = parser.parse(expr_formula)
    for i in range(100):
        df_res = parser.evaluate({'signal': df_value, 'close': df_close}, df_ann, date_arr)
    
    # -------------------------------------------------------------------------------------
    sec = '600000.SH'
    """
    # print to validate results
    print "\n======Expression Formula:\n{:s}".format(expr_formula)
    
    print "\n======Report date, ann_date and evaluation value:"
    tmp = pd.concat([df_ann.loc[:, sec], df_value.loc[:, sec]], axis=1)
    tmp.columns = ['df_ann', 'df_value']
    print tmp
    
    print "\n======Selection of result of expansion:"
    print "20161028  {:.4f}".format(df_res.loc[20161028, sec])
    print "20161031  {:.4f}".format(df_res.loc[20161031, sec])
    print "20170427  {:.4f}".format(df_res.loc[20170427, sec])
    
    """
    assert abs(df_res.loc[20161028, sec] - 82172000000) < 1
    assert abs(df_res.loc[20161031, sec] - 120928000000) < 1
    assert abs(df_res.loc[20170427, sec] - 42360000000) < 1


if __name__ == "__main__":
    import time
    t_start = time.time()
    
    test_align()
    
    t3 = time.time() - t_start
    print "\n\n\nTime lapsed in total: {:.1f}".format(t3)



# encoding: utf-8
import json
import os
import errno
import cPickle as pickle
import codecs

from .. import SOURCE_ROOT_DIR


def create_dir(filename):
    """
    Create dir if directory of filename does not exist.

    Parameters
    ----------
    filename : str

    """
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise


def read_json(fp):
    """
    Read JSON file to dict. Return None if file not found.

    Parameters
    ----------
    fp : str
        Path of the JSON file.

    Returns
    -------
    dict

    """
    content = dict()
    try:
        with codecs.open(fp, 'r', encoding='utf-8') as f:
            content = json.load(f)
    except IOError as e:
        if e.errno not in (errno.ENOENT, errno.EISDIR, errno.EINVAL):
            raise
    return content


def save_json(serializable, file_name):
    """
    Save an serializable object to JSON file.

    Parameters
    ----------
    serializable : object
    file_name : str

    """
    fn = os.path.abspath(file_name)
    create_dir(fn)
    
    with codecs.open(fn, 'w', encoding='utf-8') as f:
        json.dump(serializable, f)


def load_pickle(fp):
    """
    Read Pickle file. Return None if file not found.

    Parameters
    ----------
    fp : str
        Path of the Pickle file.

    Returns
    -------
    object or None

    """
    content = None
    try:
        with open(fp, 'rb') as f:
            content = pickle.load(f)
    except IOError as e:
        if e.errno not in (errno.ENOENT, errno.EISDIR, errno.EINVAL):
            raise
    return content


def save_pickle(obj, file_name):
    """
    Save an object to Pickle file.

    Parameters
    ----------
    obj : object
    file_name : str

    """
    fn = os.path.abspath(file_name)
    create_dir(fn)

    with open(fn, 'wb') as f:
        pickle.dump(obj, f)


def join_relative_path(*paths):
    """Get absolute path using paths that are relative to project root."""
    return os.path.abspath(os.path.join(SOURCE_ROOT_DIR, *paths))


def fig2base64(fig, format='png'):
    """
    
    Parameters
    ----------
    fig : matplotlib.fig.Figure
    format : str
        Eg. png, jpg

    Returns
    -------

    """
    import cStringIO
    import base64
    str_io = cStringIO.StringIO()
    fig.savefig(str_io, format=format)
    str_io.seek(0)
    res = base64.b64encode(str_io.read())
    return res



# encoding: utf-8

import datetime

from jaqs.data import RemoteDataService
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)


def test_calendar():
    ds = RemoteDataService()
    ds.init_from_config(data_config)

    res1 = ds.get_trade_date_range(20121224, 20130201)
    assert len(res1) == 27

    res2 = ds.get_next_trade_date(20170102)
    assert res2 == 20170103

    res3 = ds.get_next_trade_date(20170104)
    assert res3 == 20170105

    res11 = ds.get_trade_date_range(20161224, 20170201)
    assert len(res11) == 23

    assert not ds.is_trade_date(20150101)
    assert not ds.is_trade_date(20130501)


def test_dtutil():
    date = 20170808
    assert jutil.get_next_period_day(20170831, 'day', extra_offset=1) == 20170904
    assert jutil.get_next_period_day(20170831, 'day', n=2, extra_offset=0) == 20170904
    assert jutil.get_next_period_day(20170831, 'day', n=7, extra_offset=0) == 20170911
    assert jutil.get_next_period_day(20170831, 'week', extra_offset=1) == 20170905
    assert jutil.get_next_period_day(20170831, 'month', extra_offset=0) == 20170901
    
    monthly = 20170101
    while monthly < 20180301:
        monthly = jutil.get_next_period_day(monthly, 'month', extra_offset=0)
        assert datetime.datetime.strptime(str(monthly), "%Y%m%d").weekday() < 5


if __name__ == "__main__":
    test_calendar()
    test_dtutil()



# encoding: utf-8
from jaqs.data.basic import Bar, InstManager
from jaqs.data import RemoteDataService
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)


def test_inst_manager():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    
    mgr = InstManager(ds)
    mgr.load_instruments()
    sym = '000001.SZ'
    inst_obj = mgr.get_instrument(sym)
    assert inst_obj.market == 'SZ'
    assert inst_obj.symbol == sym
    assert inst_obj.multiplier == 1
    assert inst_obj.inst_type == 1


def test_bar():
    from jaqs.data import RemoteDataService
    from jaqs.trade.common import QUOTE_TYPE
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    
    df_quotes, msg = ds.bar(symbol='rb1710.SHF,hc1710.SHF', start_time=200000, end_time=160000,
                                           trade_date=20170704, freq=QUOTE_TYPE.MIN)
    quotes_list = Bar.create_from_df(df_quotes)


if __name__ == "__main__":
    test_inst_manager()
    test_bar()



# encoding: UTF-8
"""
Basic data types, classes and models for trade.

"""

from .backtest import AlphaBacktestInstance, EventBacktestInstance
from .portfoliomanager import PortfolioManager
from .realtime import EventRealTimeInstance
from .strategy import Strategy, AlphaStrategy, EventDrivenStrategy
from .tradegateway import BaseTradeApi, RealTimeTradeApi, AlphaTradeApi, BacktestTradeApi


__all__ = ['AlphaBacktestInstance', 'EventBacktestInstance',
           'PortfolioManager',
           'EventRealTimeInstance',
           'Strategy', 'AlphaStrategy', 'EventDrivenStrategy',
           'BaseTradeApi', 'RealTimeTradeApi', 'AlphaTradeApi', 'BacktestTradeApi']



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Author: AxiaCore S.A.S. http://axiacore.com
#
# Based on js-expression-eval, by Matthew Crumley (email@matthewcrumley.com, http://silentmatt.com/)
# https://github.com/silentmatt/js-expression-eval
#
# Ported to Python and modified by Vera Mazhuga (ctrl-alt-delete@live.com, http://vero4ka.info/)
#
# You are free to use and modify this code in anyway you find useful. Please leave this comment in the code
# to acknowledge its original source. If you feel like it, I enjoy hearing about projects that use my code,
# but don't feel like you have to let me know or ask permission.

# modified by symbol from quantOS.org

from __future__ import division

import math

import numpy as np
import pandas as pd

from jaqs.data.align import align
import jaqs.util.numeric as numeric

TNUMBER = 0
TOP1 = 1
TOP2 = 2
TVAR = 3
TFUNCALL = 4


class Expression(object):
    
    def __init__(self, tokens, ops1, ops2, functions):
        self.tokens = tokens
        self.ops1 = ops1
        self.ops2 = ops2
        self.functions = functions
        self.ann_dts = None
        self.trade_dts = None
        self.index_member = None
    
    def simplify(self, values):
        values = values or {}
        nstack = []
        newexpression = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == TNUMBER:
                nstack.append(item)
            elif type_ == TVAR and item.index_ in values:
                item = Token(TNUMBER, 0, 0, values[item.index_])
                nstack.append(item)
            elif type_ == TOP2 and len(nstack) > 1:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = self.ops2[item.index_]
                item = Token(TNUMBER, 0, 0, f(n1.number_, n2.number_))
                nstack.append(item)
            elif type_ == TOP1 and nstack:
                n1 = nstack.pop()
                f = self.ops1[item.index_]
                item = Token(TNUMBER, 0, 0, f(n1.number_))
                nstack.append(item)
            else:
                while len(nstack) > 0:
                    newexpression.append(nstack.pop(0))
                newexpression.append(item)
        while nstack:
            newexpression.add(nstack.pop(0))
        
        return Expression(newexpression, self.ops1, self.ops2, self.functions)
    
    def substitute(self, variable, expr):
        if not isinstance(expr, Expression):
            pass  # expr = Parser().parse(str(expr))
        newexpression = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == TVAR and item.index_ == variable:
                for j in range(0, len(expr.tokens)):
                    expritem = expr.tokens[j]
                    replitem = Token(
                            expritem.type_,
                            expritem.index_,
                            expritem.prio_,
                            expritem.number_,
                    )
                    newexpression.append(replitem)
            else:
                newexpression.append(item)
        
        ret = Expression(newexpression, self.ops1, self.ops2, self.functions)
        return ret
    
    def evaluate(self, values, ann_dts=None, trade_dts=None):
        self.ann_dts = ann_dts
        self.trade_dts = trade_dts
        values = values or {}
        nstack = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == TNUMBER:
                nstack.append(item.number_)
            elif type_ == TOP2:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = self.ops2[item.index_]
                nstack.append(f(n1, n2))
            elif type_ == TVAR:
                if item.index_ in values:
                    nstack.append(values[item.index_])
                elif item.index_ in self.functions:
                    nstack.append(self.functions[item.index_])
                else:
                    raise Exception('undefined variable: ' + item.index_)
            elif type_ == TOP1:
                n1 = nstack.pop()
                f = self.ops1[item.index_]
                nstack.append(f(n1))
            elif type_ == TFUNCALL:
                n1 = nstack.pop()
                f = nstack.pop()
                if f.apply and f.call:
                    if type(n1) is list:
                        nstack.append(f.apply(None, n1))
                    else:
                        nstack.append(f.call(None, n1))
                else:
                    raise Exception(f + ' is not a function')
            else:
                raise Exception('invalid Expression')
        if len(nstack) > 1:
            raise Exception('invalid Expression (parity)')
        return nstack[0]
    
    def toString(self, toJS=False):
        nstack = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == TNUMBER:
                nstack.append(item.number_)
            elif type_ == TOP2:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = item.index_
                if toJS and f == '^':
                    nstack.append('math.pow(' + n1 + ',' + n2 + ')')
                else:
                    nstack.append('(' + n1 + f + n2 + ')')
            elif type_ == TVAR:
                nstack.append(item.index_)
            elif type_ == TOP1:
                n1 = nstack.pop()
                f = item.index_
                if f == '-':
                    nstack.append('({0}{1})'.format(f, n1))
                else:
                    nstack.append('{0}({1})'.format(f, n1))
            elif type_ == TFUNCALL:
                n1 = nstack.pop()
                f = nstack.pop()
                nstack.append(f + '(' + n1 + ')')
            else:
                raise Exception('invalid Expression')
        if len(nstack) > 1:
            raise Exception('invalid Expression (parity)')
        return nstack[0]
    
    def variables(self):
        vars = []
        for i in range(0, len(self.tokens)):
            item = self.tokens[i]
            if item.type_ == TVAR and \
                    not item.index_ in vars and \
                    not self.functions.has_key(item.index_):
                vars.append(item.index_)
        return vars


class Token(object):
    def __init__(self, type_, index_, prio_, number_):
        self.type_ = type_
        self.index_ = index_ or 0
        self.prio_ = prio_ or 0
        self.number_ = number_ if number_ != None else 0
    
    def to_str(self):
        if self.type_ == TNUMBER:
            return self.number_
        if self.type_ == TOP1 or self.type_ == TOP2 or self.type_ == TVAR:
            return self.index_
        elif self.type_ == TFUNCALL:
            return 'CALL'
        else:
            return 'Invalid Token'


class Parser(object):
    def __init__(self):
        self.success = False
        self.errormsg = ''
        self.expression = ''
        
        self.pos = 0
        
        self.tokens = None
        self.tokennumber = 0
        self.tokenprio = 0
        self.tokenindex = 0
        self.tmpprio = 0
        
        self.PRIMARY = 1
        self.OPERATOR = 2
        self.FUNCTION = 4
        self.LPAREN = 8
        self.RPAREN = 16
        self.COMMA = 32
        self.SIGN = 64
        self.CALL = 128
        self.NULLARY_CALL = 256
        
        # do not need parenthesis
        self.ops1 = {
            'Sin': np.sin,
            'Cos': np.cos,
            'Tan': np.tan,
            #             'asin': np.asin,
            #             'acos': np.acos,
            #             'atan': np.atan,
            #            'Mean':         np.mean,
            'Sqrt': np.sqrt,
            'Log': np.log,
            'Abs': np.abs,
            'Ceil': np.ceil,
            'Floor': np.floor,
            'Round': np.round,
            '-': self.neg,
            '!': self.logicalNot,
            'Sign': np.sign,
            #            'Rank':         self.rank,
            'exp': np.exp
        }
        
        self.ops2 = {
            '+': self.add,
            '-': self.sub,
            '*': self.mul,
            '/': self.div,
            '%': self.mod,
            '^': np.power,
            ',': self.append,
            # '||': self.concat,
            "==": self.equal,
            "!=": self.notEqual,
            ">": self.greaterThan,
            "<": self.lessThan,
            ">=": self.greaterThanEqual,
            "<=": self.lessThanEqual,
            "&&": self.andOperator,
            "||": self.orOperator
        }

        # need parenthesis
        self.functions = {
            # cross section
            'Min': np.minimum,
            'Max': np.maximum,
            'Rank': self.rank,
            'Quantile': self.to_quantile,
            'Ts_Quantile': self.ts_quantile,
            'GroupQuantile': self.group_quantile,
            'GroupRank': self.group_rank,
            'ConditionRank': self.cond_rank,
            'Standardize': self.standardize,
            'Cutoff': self.cutoff,
            # 'GroupApply': self.group_apply,
            # time series
            'Ts_Rank': self.ts_rank,
            'Ewma': self.ewma,
            'Sma':self.sma,
            'Sum': self.sum,
            'Product': self.product,  # rolling product
            'CountNans': self.count_nans,  # rolling count Nans
            'StdDev': self.std_dev,
            'Covariance': self.cov,
            'Correlation': self.corr,
            'Corr': self.corr,
            'Delay': self.delay,
            'Delta': self.delta,
            'Return': self.calc_return,
            'Ts_Mean': self.ts_mean,
            'Ts_Min': self.ts_min,
            'Ts_Max': self.ts_max,
            'Ts_Skewness': self.ts_skew,
            'Ts_Kurtosis': self.ts_kurt,
            'Tail': self.tail,
            'Step': self.step,
            'Decay_linear': self.decay_linear,
            'Decay_exp': self.decay_exp,
            # inplace
            'Pow': np.power,
            'SignedPower': self.signed_power,
            # others
            'If': self.ifFunction,
            # test
        }
        
        self.consts = {
            'E': math.e,
            'PI': math.pi,
        }
        
        # no use
        self.values = {
            'sin': math.sin,
            'cos': math.cos,
            'tan': math.tan,
            'asin': math.asin,
            'acos': math.acos,
            'atan': math.atan,
            'sqrt': math.sqrt,
            'log': math.log,
            'abs': abs,
            'ceil': math.ceil,
            'floor': math.floor,
            'round': round,
            'random': self.random,
            'fac': self.fac,
            'exp': math.exp,
            'min': min,
            'max': max,
            'pyt': self.pyt,
            'pow': math.pow,
            'atan2': math.atan2,
            'E': math.e,
            'PI': math.pi
        }
        
        self.ann_dts = None
        self.trade_dts = None
    
    # -----------------------------------------------------
    # functions
    def add(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        return a + b
    
    def sub(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        return a - b
    
    def mul(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        return a * b
    
    def div(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        res = a / b
        if isinstance(res, pd.DataFrame):
            res = res.replace([np.inf, -np.inf], np.nan)
        return res
    
    def mod(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        return a % b
    
    def pow(self, a, b):
        return np.power(a, b)
    
    def concat(self, a, b, *args):
        result = u'{0}{1}'.format(a, b)
        for arg in args:
            result = u'{0}{1}'.format(result, arg)
        return result
    
    @staticmethod
    def _to_array(x):
        if isinstance(x, (pd.DataFrame, pd.Series)):
            return x.values
        elif isinstance(x, np.ndarray):
            return x
        elif isinstance(x, (int, float, bool, np.integer, np.float, np.bool)):
            return np.asarray(x)
    
    def equal(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        # arr, brr = a.values, b.values
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr == brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def notEqual(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr != brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def greaterThan(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr > brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def lessThan(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr < brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def greaterThanEqual(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr >= brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def lessThanEqual(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = arr <= brr
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def andOperator(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = np.logical_and(arr, brr)
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def orOperator(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        arr, brr = self._to_array(a), self._to_array(b)
        mask = np.logical_or(np.isnan(arr), np.isnan(brr))
        res = np.logical_or(arr, brr)
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def neg(self, a):
        return -a
    
    def logicalNot(self, a):
        arr  = self._to_array(a)
        mask = np.isnan(a)        
        res = np.logical_not(a)
        res = res.astype(float)
        res[mask] = np.nan
        return pd.DataFrame(index=a.index, columns=a.columns, data=res)
    
    def random(self, a):
        return np.random.rand() * (a or 1)
    
    def fac(self, a):  # a!
        return np.math.factorial(a)
    
    def pyt(self, a, b):
        (a, b) = self._align_bivariate(a, b)
        return np.sqrt(a * a + b * b)
    
    def ifFunction(self, cond, b, c):
        mask = np.isnan(cond)
        data = np.where(cond, b, c)
        data = data.astype(float)
        data[mask] = np.nan
        df = pd.DataFrame(data, columns=cond.columns, index=cond.index)
        return df
    
    def tail(self, x, lower, upper, neweval):
        data = np.where((x >= lower) & (x <= upper), neweval, x)
        df = pd.DataFrame(data, columns=x.columns, index=x.index)
        return df

    def append(self, a, b):
        if type(a) != list:
            return [a, b]
        a.append(b)
        return a

    # -----------------------------------------------------
    # Time Series functions. For two parameter, must align
    @staticmethod
    def ewma(df, halflife):
        r = df.ewm(halflife=halflife, axis=0)
        return r.mean()
    
    @staticmethod
    def sma(df, n, m):
        a = n * 1.0 / m - 1
        r = df.ewm(com=a, axis=0)
        return r.mean()
    
    def corr(self, x, y, n):
        (x, y) = self._align_bivariate(x, y)
        return pd.rolling_corr(x, y, n)
    
    def cov(self, x, y, n):
        (x, y) = self._align_bivariate(x, y)
        return pd.rolling_cov(x, y, n)
    
    def std_dev(self, x, n):
        return pd.rolling_std(x, n)
    
    def sum(self, x, n):
        return pd.rolling_sum(x, n)
    
    def count_nans(self, x, n):
        return n - pd.rolling_count(x, n)
    
    def delay(self, x, n):
        return x.shift(n)
    
    def delta(self, x, n):
        return x.diff(n)
    
    @staticmethod
    def calc_return(df, forward=1, log=False):
        if log:
            res = np.log(df).diff(forward)
        else:
            shift = df.shift(forward)
            res = (df - shift) / shift
        return res
    
    def ts_mean(self, x, n):
        return pd.rolling_mean(x, n)
    
    def ts_min(self, x, n):
        return pd.rolling_min(x, n)
    
    def ts_max(self, x, n):
        return pd.rolling_max(x, n)
    
    def ts_kurt(self, x, n):
        return pd.rolling_kurt(x, n)
    
    def ts_skew(self, x, n):
        return pd.rolling_skew(x, n)
    
    def product(self, x, n):
        return pd.rolling_apply(x, n, np.product)

    @staticmethod
    def ts_rank(df, window):
        """Return a DataFrame with values ranging from 0.0 to 1.0"""
        roll = df.rolling(window=window)
    
        def _rank_arr(arr, norm=1.0):
            norm = norm * 1.0
            ranks = np.argsort(np.argsort(arr))[-1] + 1
            return ranks / norm
    
        res = roll.apply(_rank_arr, kwargs={'norm': window})
        return res

    def step(self, x, n):
        st = x.copy()
        n = n + 1
        begin = n - len(x.index)
        for col in st.columns:
            st.loc[:, col] = range(begin, n, 1)
        return st
    
    def decay_exp_array(self, x, f):
        n = len(x)
        step = range(0, n)
        step = step[::-1]
        fs = np.power(f, step)
        return np.dot(x, fs) / np.sum(fs)
    
    def decay_linear_array(self, x):
        n = len(x) + 1
        step = range(1, n)
        return np.dot(x, step) / np.sum(step)
    
    def decay_linear(self, x, n):
        return pd.rolling_apply(x, n, self.decay_linear_array)
    
    def decay_exp(self, x, f, n):
        return pd.rolling_apply(x, n, self.decay_exp_array, args=[f])
    
    def signed_power(self, x, e):
        signs = np.sign(x)
        return signs * np.power(np.abs(x), e)

    # -----------------------------------------------------
    # Cross Section functions
    def cond_rank(self, x, group):
        x = self._align_univariate(x)
        g_rank = x[group]
        return g_rank.rank(axis=1)

    # -----------------------------------------------------
    # cross section functions
    def _mask_non_index_member(self, df):
        if self.index_member is not None:
            self.index_member = self.index_member.astype(bool)
            df[~self.index_member] = np.nan
        return df

    def rank(self, df):
        """Return a DataFrame with values ranging from 0.0 to 1.0"""
        df = self._align_univariate(df)
        df = self._mask_non_index_member(df)
        return df.rank(axis=1).div((df.shape[1] - df.isnull().sum(axis=1)), axis=0)

    # TODO: all cross-section operations support in-group modification: neutral, extreme values, standardize.
    def group_rank(self, x, group):
        x = self._align_univariate(x)
        x = self._mask_non_index_member(x)
        vals = np.unique(group.values.flatten())
        res = None
        for val in vals:
            rank = x[group == val].rank(axis=1, na_option='keep')
            if res is None:
                res = rank
            else:
                res = res.fillna(rank)
        return res
    
    def ts_quantile(self, df, window=3, n_quantiles=5):
        roll = df.rolling(window=window)
    
        func = lambda arr: numeric.quantilize_without_nan(arr, n_quantiles=n_quantiles, axis=0)[-1]
        res = roll.apply(func)
        return res
    
    def to_quantile(self, df, n_quantiles=5, axis=1):
        """
        Convert cross-section values to the quantile number they belong.
        Small values get small quantile numbers.
        
        Parameters
        ----------
        df : DataFrame
            index date, column symbols
        n_quantiles : int
            The number of quantile to be divided to.
        axis : int
            Axis to apply quantilize.

        Returns
        -------
        res : DataFrame
            index date, column symbols

        """
        df = self._align_univariate(df)
        df = self._mask_non_index_member(df)
        # TODO: unnecesssary warnings
        # import warnings
        # warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='py_exp')
        res_arr = numeric.quantilize_without_nan(df.values, n_quantiles=n_quantiles, axis=axis)
        res = pd.DataFrame(index=df.index, columns=df.columns, data=res_arr)
        return res

    def group_quantile(self, df, group, n_quantiles=5):
        df = self._align_univariate(df)
        df = self._mask_non_index_member(df)
        
        res = None
        for val in np.unique(group.values.flatten()):
            val_res = self.to_quantile(df[group == val], n_quantiles=n_quantiles)
            if res is None:
                res = val_res
            else:
                res = res.fillna(val_res)
        return res

    '''
        def group_apply(self, func, df_arg, *args, **kwargs):
        """
        Group on cross section (axis=1). Rank, Mean, Std, Max, Min, Standardize, cutoff.
        Single parameter.
        
        Parameters
        ----------
        func : callable
            Single parameter
        df_arg : pd.DataFrame
            The single argument of func.
            index is date, column is symbol.

        Returns
        -------
        res : pd.DataFrame

        """
        df_group = self.df_group
        
        def gp_apply(df_value, df_group_):
            """df has date index and symbol columns."""
            gp = df_value.groupby(by=df_group_, axis=1)
            res_apply = gp.apply(func, *args, **kwargs)
            return res_apply

        # align for quarterly data
        df_arg = self._align_univariate(df_arg)
        
        # validity check
        if isinstance(df_group, pd.DataFrame):
            if df_group.shape[0] == 1 or df_group.shape[1] == 1:
                df_group = df_group.squeeze()
                return gp_apply(df_arg, df_group)
            else:
                pass
        elif isinstance(df_group, pd.Series):
            return gp_apply(df_arg, df_group)
        else:
            raise NotImplementedError("type of df_group{}".format(type(df_group)))
    
        # for time-variant industry classification, we have to loop
        res_list = []
        for idx in df_arg.index:
            row = df_arg.loc[[idx], :]  # must be DataFrame, because func has certain operation axis
            row_group = df_group.loc[idx, :]  # must be Series, because groupby only support series
            tmp = gp_apply(row, row_group)
            res_list.append(tmp)
            
        res = pd.concat(res_list, axis=0)
        return res

    '''
    def standardize(self, df):
        """Cross section."""
        df = self._align_univariate(df)
        df = self._mask_non_index_member(df)
        
        axis = 1
        mean = df.mean(axis=axis)
        std = df.std(axis=axis)
        return df.sub(mean, axis=0).div(std, axis=0)
    
    def cutoff(self, df, z_score=3.0):
        """
        Cut off extreme values using Median Absolute Deviation
        
        Parameters
        ----------
        df : pd.DataFrame

        Returns
        -------
        pd.DataFrame

        """
        df = self._align_univariate(df)
        df = self._mask_non_index_member(df)

        axis = 1
        x = df.values
        
        median = np.nanmedian(x, axis=axis).reshape(-1, 1)
        diff = x - median
        diff_abs = np.abs(diff)
        mad = np.nanmedian(diff_abs, axis=axis).reshape(-1, 1)
        
        mask = diff_abs > z_score * mad
        x[mask] = 0
        x = x + z_score * mad * np.sign(diff * mask) + mask * median
        
        return pd.DataFrame(index=df.index, columns=df.columns, data=x)
    
    def industry_netural(self, x, group):
        pass
    
    # -----------------------------------------------------
    # align functions
    def _align_bivariate(self, df1, df2, force_align=False):
        if isinstance(df1, pd.DataFrame) and isinstance(df2, pd.DataFrame):
            len1 = len(df1.index)
            len2 = len(df2.index)
            if (self.ann_dts is not None) and (self.trade_dts is not None):
                if len1 > len2:
                    df2 = align(df2, self.ann_dts, self.trade_dts)
                elif len1 < len2:
                    df1 = align(df1, self.ann_dts, self.trade_dts)
                elif force_align:
                    df1 = align(df1, self.ann_dts, self.trade_dts)
                    df2 = align(df2, self.ann_dts, self.trade_dts)
        return (df1, df2)

    def _align_univariate(self, df1):
        if isinstance(df1, pd.DataFrame):
            if (self.ann_dts is not None) and (self.trade_dts is not None):
                len1 = len(df1.index)
                len2 = len(self.trade_dts)
                if len1 != len2:
                    return align(df1, self.ann_dts, self.trade_dts)
        return df1

    # -----------------------------------------------------
    # helper methods
    def set_capital(self, style='camel'):
        """
        Set capital style of function names.
        
        Parameters
        ----------
        style : {'upper', 'lower'}
            upper for 'Rank', lower for 'rank'
        
        """
        
        def set_dic_key_capital(dic, style='camel'):
            """
            
            Parameters
            ----------
            dic : dict
            style : {'upper', 'lower'}
                upper for 'Rank', lower for 'rank'

            Returns
            -------
            dict

            """
            if style == 'camel':
                # TODO: not implement
                # deli = '_'
                # res = {deli.join(s.title() for s in k.split(deli)): v for k, v in dic.viewitems()}
                res = dic
            elif style == 'lower':
                res = {k.lower(): v for k, v in dic.viewitems()}
            else:
                raise NotImplementedError("style = {}".format(style))
            return res
        
        self.functions = set_dic_key_capital(self.functions, style=style)
        self.ops1 = set_dic_key_capital(self.ops1, style=style)
    
    def register_function(self, name, func):
        """Register a new function to function map.
        
        Parameters
        ----------
        name : str
        func : callable
        
        """
        if name in self.functions:
            print "Register function failed: name [{:s}] already exist. Try another name.".format(name)
            return
        
        self.functions[name] = func

    # -----------------------------------------------------
    # parse and evaluate
    def parse(self, expr):
        """
        Parse a string expression.
        
        Parameters
        ----------
        expr : str
            Format of expr should follow our document.

        Returns
        -------
        Expression

        """
        self.errormsg = ''
        self.success = True
        operstack = []
        tokenstack = []
        self.tmpprio = 0
        expected = self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
        noperators = 0
        self.expression = expr
        self.pos = 0
        
        while self.pos < len(self.expression):
            if self.is_operator():
                if self.isSign() and expected & self.SIGN:
                    if self.isNegativeSign():
                        self.tokenprio = 5
                        self.tokenindex = '-'
                        noperators += 1
                        self.addfunc(tokenstack, operstack, TOP1)                    
                    elif self.isLogicNot():
                        self.tokenprio = 5
                        self.tokenindex = '!'
                        noperators += 1
                        self.addfunc(tokenstack, operstack, TOP1)
                    expected = \
                        self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
                elif self.isComment():
                    pass
                else:
                    if expected and self.OPERATOR == 0:
                        self.error_parsing(self.pos, 'unexpected operator')
                    noperators += 2
                    self.addfunc(tokenstack, operstack, TOP2)
                    expected = \
                        self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
            elif self.is_number():
                if expected and self.PRIMARY == 0:
                    self.error_parsing(self.pos, 'unexpected number')
                token = Token(TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(token)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.is_str():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected string')
                token = Token(TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(token)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.isLeftParenth():
                if (expected & self.LPAREN) == 0:
                    self.error_parsing(self.pos, 'unexpected \"(\"')
                if expected & self.CALL:
                    noperators += 2
                    self.tokenprio = -2
                    self.tokenindex = -1
                    self.addfunc(tokenstack, operstack, TFUNCALL)
                expected = \
                    self.PRIMARY | self.LPAREN | self.FUNCTION | \
                    self.SIGN | self.NULLARY_CALL
            elif self.isRightParenth():
                if expected & self.NULLARY_CALL:
                    token = Token(TNUMBER, 0, 0, [])
                    tokenstack.append(token)
                elif (expected & self.RPAREN) == 0:
                    self.error_parsing(self.pos, 'unexpected \")\"')
                expected = \
                    self.OPERATOR | self.RPAREN | self.COMMA | \
                    self.LPAREN | self.CALL
            elif self.isComma():
                if (expected & self.COMMA) == 0:
                    self.error_parsing(self.pos, 'unexpected \",\"')
                self.addfunc(tokenstack, operstack, TOP2)
                noperators += 2
                expected = \
                    self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
            elif self.is_const():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected constant')
                consttoken = Token(TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(consttoken)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.isOp2():
                if (expected & self.FUNCTION) == 0:
                    self.error_parsing(self.pos, 'unexpected function')
                self.addfunc(tokenstack, operstack, TOP2)
                noperators += 2
                expected = self.LPAREN
            elif self.isOp1():
                if (expected & self.FUNCTION) == 0:
                    self.error_parsing(self.pos, 'unexpected function')
                self.addfunc(tokenstack, operstack, TOP1)
                noperators += 1
                expected = self.LPAREN
            elif self.isVar():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected variable')
                
                vartoken = Token(TVAR, self.tokenindex, 0, 0)
                tokenstack.append(vartoken)
                expected = \
                    self.OPERATOR | self.RPAREN | \
                    self.COMMA | self.LPAREN | self.CALL
            elif self.isWhite():
                pass
            else:
                if self.errormsg == '':
                    self.error_parsing(self.pos, "You have an unknown character in your formula;"
                                                 "at '{:s}'".format(self.expression[self.pos: 10]))
                else:
                    self.error_parsing(self.pos, self.errormsg)
        if self.tmpprio < 0 or self.tmpprio >= 10:
            self.error_parsing(self.pos, 'unmatched \"()\"')
        while len(operstack) > 0:
            tmp = operstack.pop()
            tokenstack.append(tmp)
        if (noperators + 1) != len(tokenstack):
            self.error_parsing(self.pos, 'parity')
        self.tokens = tokenstack
        return Expression(tokenstack, self.ops1, self.ops2, self.functions)
    
    def evaluate(self, values, ann_dts=None, trade_dts=None, index_member=None):
        """
        Evaluate the value of expression using. Data of different frequency will be automatically expanded.
        
        Parameters
        ----------
        values : dict
            Key is variable name, value is pd.DataFrame (index is date, column is symbol)
        ann_dts : pd.DataFrame
            Announcement dates of financial statements of securities.
        trade_dts : np.ndarray
            The date index of result.
        index_member : pd.DataFrame

        Returns
        -------
        pd.DataFrame

        """
        self.ann_dts = ann_dts
        self.trade_dts = trade_dts
        self.index_member = index_member
        
        values = values or {}
        nstack = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == TNUMBER:
                nstack.append(item.number_)
            elif type_ == TOP2:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = self.ops2[item.index_]
                nstack.append(f(n1, n2))
            elif type_ == TVAR:
                if item.index_ in values:
                    nstack.append(values[item.index_])
                elif item.index_ in self.functions:
                    nstack.append(self.functions[item.index_])
                else:
                    raise Exception('undefined variable: ' + item.index_)
            elif type_ == TOP1:
                n1 = nstack.pop()
                f = self.ops1[item.index_]
                nstack.append(f(n1))
            elif type_ == TFUNCALL:
                n1 = nstack.pop()
                f = nstack.pop()
                if callable(f):
                    if type(n1) is list:
                        nstack.append(f(*n1))
                    else:
                        nstack.append(f(n1))  # call(f, n1)
                else:
                    raise Exception(f + ' is not a function')
            else:
                raise Exception('invalid Expression')
        if len(nstack) > 1:
            raise Exception('invalid Expression (parity)')
        return nstack[0]

    # -----------------------------------------------------
    # Other
    def error_parsing(self, column, msg):
        self.success = False
        self.errormsg = 'parse error [column ' + str(column) + ']: ' + msg
        raise Exception(self.errormsg)
    
    def addfunc(self, tokenstack, operstack, type_):
        operator = Token(
                type_,
                self.tokenindex,
                self.tokenprio + self.tmpprio,
                0,
        )
        while len(operstack) > 0:
            if operator.prio_ <= operstack[len(operstack) - 1].prio_:
                tokenstack.append(operstack.pop())
            else:
                break
        operstack.append(operator)
    
    def is_number(self):
        r = False
        str = ''
        while self.pos < len(self.expression):
            code = self.expression[self.pos]
            if (code >= '0' and code <= '9') or code == '.':
                str += self.expression[self.pos]
                self.pos += 1
                if '.' in str:
                    self.tokennumber = float(str)
                else:
                    self.tokennumber = int(str)
                r = True
            else:
                break
        return r
    
    def unescape(self, v, pos):
        buffer = []
        escaping = False
        
        for i in range(0, len(v)):
            c = v[i]
            
            if escaping:
                if c == "'":
                    buffer.append("'")
                    break
                elif c == '\\':
                    buffer.append('\\')
                    break
                elif c == '/':
                    buffer.append('/')
                    break
                elif c == 'b':
                    buffer.append('\b')
                    break
                elif c == 'f':
                    buffer.append('\f')
                    break
                elif c == 'n':
                    buffer.append('\n')
                    break
                elif c == 'r':
                    buffer.append('\r')
                    break
                elif c == 't':
                    buffer.append('\t')
                    break
                elif c == 'u':
                    # interpret the following 4 characters
                    # as the hex of the unicode code point
                    codePoint = int(v[i + 1, i + 5], 16)
                    buffer.append(unichr(codePoint))
                    i += 4
                    break
                else:
                    raise self.error_parsing(
                            pos + i,
                            'Illegal escape sequence: \'\\' + c + '\'',
                    )
                escaping = False
            else:
                if c == '\\':
                    escaping = True
                else:
                    buffer.append(c)
        
        return ''.join(buffer)
    
    def is_str(self):
        r = False
        str = ''
        startpos = self.pos
        if self.pos < len(self.expression) and self.expression[self.pos] == "'":
            self.pos += 1
            while self.pos < len(self.expression):
                code = self.expression[self.pos]
                if code != '\'' or (str != '' and str[-1] == '\\'):
                    str += self.expression[self.pos]
                    self.pos += 1
                else:
                    self.pos += 1
                    self.tokennumber = self.unescape(str, startpos)
                    r = True
                    break
        return r
    
    def is_const(self):
        for i in self.consts:
            L = len(i)
            str = self.expression[self.pos:self.pos + L]
            if i == str:
                if len(self.expression) <= self.pos + L:
                    self.tokennumber = self.consts[i]
                    self.pos += L
                    return True
                if not self.expression[self.pos + L].isalnum() and self.expression[self.pos + L] != "_":
                    self.tokennumber = self.consts[i]
                    self.pos += L
                    return True
        return False
    
    def is_operator(self):
        ops = (
            ('+', 2, '+'),
            ('-', 2, '-'),
            ('*', 3, '*'),
            (u'\u2219', 3, '*'),  # bullet operator
            (u'\u2022', 3, '*'),  # black small circle
            ('/', 4, '/'),
            ('%', 4, '%'),
            ('^', 6, '^'),
            ('&&', 1, '&&'),
            ('||', 1, '||'),
            ('==', 1, '=='),
            ('!=', 1, '!='),
            ('<=', 1, '<='),
            ('>=', 1, '>='),
            ('<', 1, '<'),
            ('>', 1, '>'),
            ('and', 0, 'and'),
            ('or', 0, 'or'),
            ('!', 5, '!'),
        )
        for token, priority, index in ops:
            if self.expression.startswith(token, self.pos):
                self.tokenprio = priority
                self.tokenindex = index
                self.pos += len(token)
                return True
        return False
    
    def isSign(self):
        code = self.expression[self.pos - 1]
        return (code == '+') or (code == '-') or (code == '!')
    
    def isPositiveSign(self):
        code = self.expression[self.pos - 1]
        return code == '+'
    
    def isNegativeSign(self):
        code = self.expression[self.pos - 1]
        return code == '-' 
    
    def isLogicNot(self):
        code = self.expression[self.pos - 1]
        return code == '!' 
    
    def isLeftParenth(self):
        code = self.expression[self.pos]
        if code == '(':
            self.pos += 1
            self.tmpprio += 10
            return True
        return False
    
    def isRightParenth(self):
        code = self.expression[self.pos]
        if code == ')':
            self.pos += 1
            self.tmpprio -= 10
            return True
        return False
    
    def isComma(self):
        code = self.expression[self.pos]
        if code == ',':
            self.pos += 1
            self.tokenprio = -1
            self.tokenindex = ","
            return True
        return False
    
    def isWhite(self):
        code = self.expression[self.pos]
        if code.isspace():
            self.pos += 1
            return True
        return False
    
    def isOp1(self):
        str = ''
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.upper() == c.lower():
                if i == self.pos or (c != '_' and (c < '0' or c > '9')):
                    break
            str += c
        if len(str) > 0 and str in self.ops1:
            self.tokenindex = str
            self.tokenprio = 7
            self.pos += len(str)
            return True
        return False
    
    def isOp2(self):
        str = ''
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.upper() == c.lower():
                if i == self.pos or (c != '_' and (c < '0' or c > '9')):
                    break
            str += c
        if len(str) > 0 and (str in self.ops2):
            self.tokenindex = str
            self.tokenprio = 7
            self.pos += len(str)
            return True
        return False
    
    def isVar(self):
        str = ''
        inQuotes = False
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.lower() == c.upper():
                if ((i == self.pos and c != '"') or (not (c in '_."') and (c < '0' or c > '9'))) and not inQuotes:
                    break
            if c == '"':
                inQuotes = not inQuotes
            str += c
        if str:
            self.tokenindex = str
            self.tokenprio = 4
            self.pos += len(str)
            return True
        return False
    
    def isComment(self):
        code = self.expression[self.pos - 1]
        if code == '/' and self.expression[self.pos] == '*':
            self.pos = self.expression.index('*/', self.pos) + 2
            if self.pos == 1:
                self.pos = len(self.expression)
            return True
        return False



# encoding: utf-8

import numpy as np
import pandas as pd

from jaqs.trade import common
from jaqs.data.basic import Bar
from jaqs.data.basic import Trade
import jaqs.util as jutil


class BacktestInstance(object):
    """
    Attributes
    ----------
    start_date : int
    end_date : int
    
    """
    def __init__(self):
        super(BacktestInstance, self).__init__()
        
        self.strategy = None
        self.start_date = 0
        self.end_date = 0

        self.props = None
        
        self.ctx = None
        
    def init_from_config(self, props):
        """
        
        Parameters
        ----------
        props : dict

        """
        for name in ['start_date', 'end_date']:
            if name not in props:
                raise ValueError("{} must be provided in props.".format(name))
        
        self.props = props
        self.start_date = props.get("start_date")
        self.end_date = props.get("end_date")
        
        if 'symbol' in props:
            self.ctx.init_universe(props['symbol'])
        elif hasattr(self.ctx, 'dataview'):
            self.ctx.init_universe(self.ctx.dataview.symbol)
        else:
            raise ValueError("No dataview, no symbol either.")

        for obj in ['data_api', 'trade_api', 'pm', 'strategy']:
            obj = getattr(self.ctx, obj)
            if obj is not None:
                obj.init_from_config(props)



'''
class AlphaBacktestInstance_OLD_dataservice(BacktestInstance):
    def __init__(self):
        BacktestInstance.__init__(self)
        
        self.last_rebalance_date = 0
        self.current_rebalance_date = 0
        self.trade_days = None
    
    def _is_trade_date(self, start, end, date, data_server):
        if self.trade_days is None:
            df, msg = data_server.daily('000300.SH', start, end, fields="close")
            self.trade_days = df.loc[:, 'trade_date'].values
        return date in self.trade_days
    
    def go_next_rebalance_day(self):
        """update self.ctx.trade_date and last_date."""
        if self.ctx.gateway.match_finished:
            next_period_day = jutil.get_next_period_day(self.ctx.trade_date,
                                                         self.ctx.strategy.period, self.ctx.strategy.days_delay)
            # update current_date: next_period_day is a workday, but not necessarily a trade date
            if self.ctx.calendar.is_trade_date(next_period_day):
                self.ctx.trade_date = next_period_day
            else:
                self.ctx.trade_date = self.ctx.calendar.get_next_trade_date(next_period_day)
            self.ctx.trade_date = self.ctx.calendar.get_next_trade_date(next_period_day)

            # update re-balance date
            if self.current_rebalance_date > 0:
                self.last_rebalance_date = self.current_rebalance_date
            else:
                self.last_rebalance_date = self.start_date
            self.current_rebalance_date = self.ctx.trade_date
        else:
            # TODO here we must make sure the matching will not last to next period
            self.ctx.trade_date = self.ctx.calendar.get_next_trade_date(self.ctx.trade_date)

        self.last_date = self.ctx.calendar.get_last_trade_date(self.ctx.trade_date)

    def run_alpha(self):
        gateway = self.ctx.gateway
        
        self.ctx.trade_date = self.start_date
        while True:
            self.go_next_rebalance_day()
            if self.ctx.trade_date > self.end_date:
                break
            
            if gateway.match_finished:
                self.on_new_day(self.last_date)
                df_dic = self.ctx.strategy.get_univ_prices()  # access data
                self.ctx.strategy.re_balance_plan_before_open(df_dic, suspensions=[])
                
                self.on_new_day(self.ctx.trade_date)
                self.ctx.strategy.send_bullets()
            else:
                self.on_new_day(self.ctx.trade_date)
            
            df_dic = self.ctx.strategy.get_univ_prices()  # access data
            trade_indications = gateway.match(df_dic, self.ctx.trade_date)
            for trade_ind in trade_indications:
                self.ctx.strategy.on_trade(trade_ind)
        
        print "Backtest done. {:d} days, {:.2e} trades in total.".format(len(self.trade_days),
                                                                         len(self.ctx.strategy.pm.trades))
    
    def on_new_day(self, date):
        self.ctx.trade_date = date
        self.ctx.strategy.on_new_day(date)
        self.ctx.gateway.on_new_day(date)
    
    def save_results(self, folder='../output/'):
        import pandas as pd
        
        trades = self.ctx.strategy.pm.trades
        
        type_map = {'task_id': str,
                    'entrust_no': str,
                    'entrust_action': str,
                    'symbol': str,
                    'fill_price': float,
                    'fill_size': int,
                    'fill_date': int,
                    'fill_time': int,
                    'fill_no': str}
        # keys = trades[0].__dict__.keys()
        ser_list = dict()
        for key in type_map.keys():
            v = [t.__getattribute__(key) for t in trades]
            ser = pd.Series(data=v, index=None, dtype=type_map[key], name=key)
            ser_list[key] = ser
        df_trades = pd.DataFrame(ser_list)
        df_trades.index.name = 'index'
        
        from os.path import join
        trades_fn = join(folder, 'trades.csv')
        configs_fn = join(folder, 'configs.json')
        fileio.create_dir(trades_fn)
        
        df_trades.to_csv(trades_fn)
        fileio.save_json(self.props, configs_fn)

        print ("Backtest results has been successfully saved to:\n" + folder)


'''


class AlphaBacktestInstance(BacktestInstance):
    """
    Backtest alpha strategy using DataView.
    
    Attributes
    ----------
    last_rebalance_date : int
    current_rebalance_date : int
    univ_price_dic : dict
        Prices of symbols at current_date

    """
    def __init__(self):
        super(AlphaBacktestInstance, self).__init__()
    
        self.last_date = 0
        self.last_rebalance_date = 0
        self.current_rebalance_date = 0
        
        self.univ_price_dic = {}
        
        self.POSITION_ADJUST_NO = 101010
        self.POSITION_ADJUST_TIME = 200000
        self.DELIST_ADJUST_NO = 202020
        self.DELIST_ADJUST_TIME = 150000
        
        self.commission_rate = 20E-4
    
    def init_from_config(self, props):
        super(AlphaBacktestInstance, self).init_from_config(props)
        
        self.commission_rate = props.get('commission_rate', 20E-4)

    def position_adjust(self):
        """
        adjust happens after market close
        Before each re-balance day, adjust for all dividend and cash paid actions during the last period.
        We assume all cash will be re-invested.
        Since we adjust our position at next re-balance day, PnL before that may be incorrect.

        """
        start = self.last_rebalance_date  # start will be one day later
        end = self.current_rebalance_date  # end is the same to ensure position adjusted for dividend on rebalance day
        df_adj = self.ctx.dataview.get_ts('adjust_factor',
                                          start_date=start, end_date=end)
        pm = self.ctx.pm
        for symbol in pm.holding_securities:
            ser = df_adj.loc[:, symbol]
            ser_div = ser.div(ser.shift(1)).fillna(1.0)
            mask_diff = ser_div != 1
            ser_adj = ser_div.loc[mask_diff]
            for date, ratio in ser_adj.iteritems():
                pos_old = pm.get_position(symbol).current_size
                # TODO pos will become float, original: int
                pos_new = pos_old * ratio
                pos_diff = pos_new - pos_old  # must be positive
                if pos_diff <= 0:
                    # TODO this is possible
                    # raise ValueError("pos_diff <= 0")
                    continue
                
                trade_ind = Trade()
                trade_ind.symbol = symbol
                trade_ind.task_id = self.POSITION_ADJUST_NO
                trade_ind.entrust_no = self.POSITION_ADJUST_NO
                trade_ind.entrust_action = common.ORDER_ACTION.BUY  # for now only BUY
                trade_ind.set_fill_info(price=0.0, size=pos_diff, date=date, time=200000, no=self.POSITION_ADJUST_NO)
                
                self.ctx.strategy.on_trade(trade_ind)

    def delist_adjust(self):
        df_inst = self.ctx.dataview.data_inst

        start = self.last_rebalance_date  # start will be one day later
        end = self.current_rebalance_date  # end is the same to ensure position adjusted for dividend on rebalance day
        
        mask = np.logical_and(df_inst['delist_date'] >= start, df_inst['delist_date'] <= end)
        dic_inst = df_inst.loc[mask, :].to_dict(orient='index')
        
        if not dic_inst:
            return
        pm = self.ctx.pm
        for symbol in pm.holding_securities.copy():
            value_dic = dic_inst.get(symbol, None)
            if value_dic is None:
                continue
            pos = pm.get_position(symbol).current_size
            last_trade_date = self._get_last_trade_date(value_dic['delist_date'])
            last_close_price = self.ctx.dataview.get_snapshot(last_trade_date, symbol=symbol, fields='close')
            last_close_price = last_close_price.at[symbol, 'close']
            
            trade_ind = Trade()
            trade_ind.symbol = symbol
            trade_ind.task_id = self.DELIST_ADJUST_NO
            trade_ind.entrust_no = self.DELIST_ADJUST_NO
            trade_ind.entrust_action = common.ORDER_ACTION.SELL  # for now only BUY
            trade_ind.set_fill_info(price=last_close_price, size=pos,
                                    date=last_trade_date, time=150000, no=self.DELIST_ADJUST_NO)
    
            self.ctx.strategy.on_trade(trade_ind)

    def re_balance_plan_before_open(self):
        """
        Do portfolio re-balance before market open (not knowing suspensions) only calculate weights.
        For now, we stick to the same close price when calculate market value and do re-balance.
        
        Parameters
        ----------

        """
        # Step.1 set weights of those non-index-members to zero
        # only filter index members when universe is defined
        universe_list = self.ctx.universe
        if self.ctx.dataview.universe:
            col = 'index_member'
            df_is_member = self.ctx.dataview.get_snapshot(self.ctx.trade_date, fields=col)
            df_is_member = df_is_member.fillna(0).astype(bool)
            dic_index_member = df_is_member.loc[:, col].to_dict()
            universe_list = [symbol for symbol, value in dic_index_member.items() if value]

        # Step.2 filter out those not listed or already de-listed
        df_inst = self.ctx.dataview.data_inst
        mask = np.logical_and(self.ctx.trade_date > df_inst['list_date'],
                              self.ctx.trade_date < df_inst['delist_date'])
        listing_symbols = df_inst.loc[mask, :].index.values
        universe_list = [s for s in universe_list if s in listing_symbols]
        
        # step.3 construct portfolio using models
        self.ctx.strategy.portfolio_construction(universe_list)
        
    def re_balance_plan_after_open(self):
        """
        Do portfolio re-balance after market open.
        With suspensions known, we re-calculate weights and generate orders.
        
        Notes
        -----
        Price here must not be adjusted.

        """
        prices = {k: v['close'] for k, v in self.univ_price_dic.viewitems()}
        # suspensions & limit_reaches: list of str
        suspensions = self.get_suspensions()
        limit_reaches = self.get_limit_reaches()
        all_list = reduce(lambda s1, s2: s1.union(s2), [set(suspensions), set(limit_reaches)])
    
        # step1. weights of those suspended and limit will be remove, and weights of others will be re-normalized
        self.ctx.strategy.re_weight_suspension(all_list)
    
        # step2. calculate market value and cash
        # market value does not include those suspended
        market_value_float, market_value_frozen = self.ctx.pm.market_value(prices, all_list)
        cash_available = self.ctx.strategy.cash + market_value_float
    
        cash_to_use = cash_available * self.ctx.strategy.position_ratio
        cash_unuse = cash_available - cash_to_use
    
        # step3. generate target positions
        # position of those suspended will remain the same (will not be traded)
        goals, cash_remain = self.ctx.strategy.generate_weights_order(self.ctx.strategy.weights, cash_to_use, prices,
                                                                  suspensions=all_list)
        self.ctx.strategy.goal_positions = goals
        self.ctx.strategy.cash = cash_remain + cash_unuse
        # self.liquidate_all()
        
        self.ctx.strategy.on_after_rebalance(cash_available + market_value_frozen)

    def run_alpha(self):
        tapi = self.ctx.trade_api
        
        self.ctx.trade_date = self._get_next_trade_date(self.start_date)
        self.last_date = self._get_last_trade_date(self.ctx.trade_date)
        while True:
            print "\n=======new day {}".format(self.ctx.trade_date)

            # match uncome orders or re-balance
            if tapi.match_finished:
                # Step1.
                # position adjust according to dividend, cash paid, de-list actions during the last period
                # two adjust must in order
                self.position_adjust()
                self.delist_adjust()

                # Step2.
                # plan re-balance before market open of the re-balance day:
                self.on_new_day(self.last_date)  # use last trade date because strategy can only access data of last day
                # get index memebers, get signals, generate weights
                self.re_balance_plan_before_open()

                # Step3.
                # do re-balance on the re-balance day
                self.on_new_day(self.ctx.trade_date)
                # get suspensions, get up/down limits, generate goal positions and send orders.
                self.re_balance_plan_after_open()
                self.ctx.strategy.send_bullets()
            else:
                self.on_new_day(self.ctx.trade_date)
            
            # Deal with trade indications
            # results = gateway.match(self.univ_price_dic)
            results = tapi.match_and_callback(self.univ_price_dic)
            for trade_ind, order_status_ind in results:
                self.ctx.strategy.cash -= trade_ind.commission
                
            self.on_after_market_close()
            
            # switch trade date
            backtest_finish = self.go_next_rebalance_day()
            if backtest_finish:
                break
        
        print "Backtest done. {:d} days, {:.2e} trades in total.".format(len(self.ctx.dataview.dates),
                                                                         len(self.ctx.pm.trades))
    
    def on_after_market_close(self):
        self.ctx.trade_api.on_after_market_close()
        
    '''
    def get_univ_prices(self, field_name='close'):
        dv = self.ctx.dataview
        df = dv.get_snapshot(self.ctx.trade_date, fields=field_name)
        res = df.to_dict(orient='index')
        return res
    
    '''
    def _is_trade_date(self, date):
        if self.ctx.dataview is not None:
            return date in self.ctx.dataview.dates
        else:
            return self.ctx.data_api.is_trade_date(date)
    
    def _get_next_trade_date(self, date):
        if self.ctx.dataview is not None:
            dates = self.ctx.dataview.dates
            mask = dates > date
            return dates[mask][0]
        else:
            return self.ctx.data_api.get_next_trade_date(date)
    
    def _get_last_trade_date(self, date):
        if self.ctx.dataview is not None:
            dates = self.ctx.dataview.dates
            mask = dates < date
            return dates[mask][-1]
        else:
            return self.ctx.data_api.get_last_trade_date(date)
    
    def go_next_rebalance_day(self):
        """
        update self.ctx.trade_date and last_date.
        
        Returns
        -------
        bool
            Whether the backtest is finished.

        """
        current_date = self.ctx.trade_date
        if self.ctx.trade_api.match_finished:
            next_period_day = jutil.get_next_period_day(current_date, self.ctx.strategy.period,
                                                        n=self.ctx.strategy.n_periods,
                                                        extra_offset=self.ctx.strategy.days_delay)
            if next_period_day > self.end_date:
                return True
            
            # update current_date: next_period_day is a workday, but not necessarily a trade date
            if self._is_trade_date(next_period_day):
                current_date = next_period_day
            else:
                try:
                    current_date = self._get_next_trade_date(next_period_day)
                except IndexError:
                    return True
        
            # update re-balance date
            if self.current_rebalance_date > 0:
                self.last_rebalance_date = self.current_rebalance_date
            else:
                self.last_rebalance_date = current_date
            self.current_rebalance_date = current_date
        else:
            # TODO here we must make sure the matching will not last to next period
            try:
                current_date = self._get_next_trade_date(current_date)
            except IndexError:
                return True
    
        self.ctx.trade_date = current_date
        self.last_date = self._get_last_trade_date(current_date)
        return False
    
    def get_suspensions(self):
        trade_status = self.ctx.dataview.get_snapshot(self.ctx.trade_date, fields='trade_status')
        trade_status = trade_status.loc[:, 'trade_status']
        # trade_status: {'N', 'XD', 'XR', 'DR', 'JiaoYi', 'TingPai', NUll (before 2003)}
        mask_sus = trade_status == u'停牌'.encode('utf-8')
        return list(trade_status.loc[mask_sus].index.values)

    def get_limit_reaches(self):
        # TODO: 10% is not the absolute value to check limit reach
        df_open = self.ctx.dataview.get_snapshot(self.ctx.trade_date, fields='open')
        df_close = self.ctx.dataview.get_snapshot(self.last_date, fields='close')
        merge = pd.concat([df_close, df_open], axis=1)
        merge.loc[:, 'limit'] = np.abs((merge['open'] - merge['close']) / merge['close']) > 9.5E-2
        return list(merge.loc[merge.loc[:, 'limit'], :].index.values)
    
    def on_new_day(self, date):
        # self.ctx.strategy.on_new_day(date)
        self.ctx.trade_api.on_new_day(date)
        
        self.ctx.snapshot = self.ctx.dataview.get_snapshot(date)
        self.univ_price_dic = self.ctx.snapshot.loc[:, ['close', 'vwap', 'open', 'high', 'low']].to_dict(orient='index')
    
    def save_results(self, folder_path='.'):
        import os
        import pandas as pd
        folder_path = os.path.abspath(folder_path)
    
        trades = self.ctx.pm.trades
    
        type_map = {'task_id': str,
                    'entrust_no': str,
                    'entrust_action': str,
                    'symbol': str,
                    'fill_price': float,
                    'fill_size': float,
                    'fill_date': int,
                    'fill_time': int,
                    'fill_no': str,
                    'commission': float}
        # keys = trades[0].__dict__.keys()
        ser_list = dict()
        for key in type_map.keys():
            v = [t.__getattribute__(key) for t in trades]
            ser = pd.Series(data=v, index=None, dtype=type_map[key], name=key)
            ser_list[key] = ser
        df_trades = pd.DataFrame(ser_list)
        df_trades.index.name = 'index'
    
        trades_fn = os.path.join(folder_path, 'trades.csv')
        configs_fn = os.path.join(folder_path, 'configs.json')
        jutil.create_dir(trades_fn)
    
        df_trades.to_csv(trades_fn)
        jutil.save_json(self.props, configs_fn)
    
        print ("Backtest results has been successfully saved to:\n" + folder_path)
    
    def show_position_info(self):
        pm = self.ctx.pm
        
        prices = {k: v['open'] for k, v in self.univ_price_dic.viewitems()}
        market_value_float, market_value_frozen = pm.market_value(prices)
        for symbol in pm.holding_securities:
            p = prices[symbol]
            size = pm.get_position(symbol).current_size
            print "{}  {:.2e}   {:.1f}@{:.2f}".format(symbol, p*size*100, p, size)
        print "float {:.2e}, frozen {:.2e}".format(market_value_float, market_value_frozen)


class EventBacktestInstance(BacktestInstance):
    def __init__(self):
        super(EventBacktestInstance, self).__init__()
        
        self.bar_type = ""
        
    def init_from_config(self, props):
        super(EventBacktestInstance, self).init_from_config(props)
        
        self.bar_type = props.get("bar_type", "1d")
        
    def go_next_trade_date(self):
        next_dt = self.ctx.data_api.get_next_trade_date(self.ctx.trade_date)
        
        self.ctx.trade_date = next_dt
    
    def on_new_day(self, date):
        self.ctx.trade_date = date
        self.ctx.time = 0
        if hasattr(self.ctx.trade_api, 'on_new_day'):
            self.ctx.trade_api.on_new_day(self.ctx.trade_date)
        if hasattr(self.ctx.trade_api, 'on_new_day'):
            self.ctx.trade_api.on_new_day(self.ctx.trade_date)
        self.ctx.strategy.initialize()
        print 'on_new_day in trade {}'.format(self.ctx.trade_date)
    
    def on_after_market_close(self):
        pass

    def _create_time_symbol_bars(self, date):
        from collections import defaultdict
        
        # query quotes data
        symbols_str = ','.join(self.ctx.universe)
        df_quotes, msg = self.ctx.data_api.bar(symbol=symbols_str, start_time=200000, end_time=160000,
                                               trade_date=date, freq=self.bar_type)
        if msg != '0,':
            print msg
        if df_quotes is None or df_quotes.empty:
            return dict()
    
        # create nested dict
        quotes_list = Bar.create_from_df(df_quotes)
        
        dic = defaultdict(dict)
        for quote in quotes_list:
            dic[jutil.combine_date_time(quote.date, quote.time)][quote.symbol] = quote
        return dic
    
    def _run_bar(self):
        """Quotes of different symbols will be aligned into one dictionary."""
        trade_dates = self.ctx.data_api.get_trade_date_range(self.start_date, self.end_date)

        for trade_date in trade_dates:
            self.on_new_day(trade_date)
            
            quotes_dic = self._create_time_symbol_bars(trade_date)
            for dt in sorted(quotes_dic.keys()):
                _, time = jutil.split_date_time(dt)
                self.ctx.time = time
                
                quote_by_symbol = quotes_dic.get(dt)
                self._process_quote_bar(quote_by_symbol)
            
            self.on_after_market_close()
    
    def _run_daily(self):
        """Quotes of different symbols will be aligned into one dictionary."""
        from collections import defaultdict
        
        symbols_str = ','.join(self.ctx.universe)
        df_daily, msg = self.ctx.data_api.daily(symbol=symbols_str, start_date=self.start_date, end_date=self.end_date,
                                                adjust_mode='post')
        if msg != '0,':
            print msg
        if df_daily is None or df_daily.empty:
            return dict()
        
        # create nested dict
        quotes_list = Bar.create_from_df(df_daily)

        dic = defaultdict(dict)
        for quote in quotes_list:
            dic[quote.trade_date][quote.symbol] = quote
        
        dates = sorted(dic.keys())
        for i in range(len(dates) - 1):
            d1, d2 = dates[i], dates[i + 1]
            self.on_new_day(d2)
            
            quote1 = dic.get(d1)
            quote2 = dic.get(d2)
            self._process_quote_daily(quote1, quote2)
            
            self.on_after_market_close()
    
    def _process_quote_daily(self, quote_yesterday, quote_today):
        # on_quote
        self.ctx.strategy.on_quote(quote_yesterday)
        
        self.ctx.trade_api.match_and_callback(quote_today, freq=self.bar_type)
        
        '''
        # match
        trade_results = self.ctx.gateway._process_quote(quote_today, freq=self.bar_type)

        # trade indication
        for trade_ind, status_ind in trade_results:
            comm = self.calc_commission(trade_ind)
            trade_ind.commission = comm
            # self.ctx.strategy.cash -= comm
            
            self.ctx.strategy.on_trade(trade_ind)
            self.ctx.strategy.on_order_status(status_ind)
        '''
        
        self.on_after_market_close()

    def run(self):
        if self.bar_type == common.QUOTE_TYPE.DAILY:
            self._run_daily()
        
        elif (self.bar_type == common.QUOTE_TYPE.MIN
              or self.bar_type == common.QUOTE_TYPE.FIVEMIN
              or self.bar_type == common.QUOTE_TYPE.QUARTERMIN):
            self._run_bar()
        
        else:
            raise NotImplementedError("bar_type = {}".format(self.bar_type))
        
        print "Backtest done."
        
    def _process_quote_bar(self, quotes_dic):
        self.ctx.trade_api.match_and_callback(quotes_dic, freq=self.bar_type)
        
        '''
        # match
        trade_results = self.ctx.trade_api._process_quote(quotes_dic, freq=self.bar_type)
        
        # trade indication
        for trade_ind, status_ind in trade_results:
            comm = self.calc_commission(trade_ind)
            trade_ind.commission = comm
            # self.ctx.strategy.cash -= comm
    
            self.ctx.strategy.on_trade(trade_ind)
            self.ctx.strategy.on_order_status(status_ind)
        '''
        
        # on_quote
        self.ctx.strategy.on_quote(quotes_dic)

    '''
    def generate_report(self, output_format=""):
        return self.pnlmgr.generateReport(output_format)
    '''
    
    def save_results(self, folder_path='.'):
        import os
        import pandas as pd
        folder_path = os.path.abspath(folder_path)
    
        trades = self.ctx.pm.trades
    
        type_map = {'task_id': str,
                    'entrust_no': str,
                    'entrust_action': str,
                    'symbol': str,
                    'fill_price': float,
                    'fill_size': float,
                    'fill_date': int,
                    'fill_time': int,
                    'fill_no': str,
                    'commission': float}
        # keys = trades[0].__dict__.keys()
        ser_list = dict()
        for key in type_map.keys():
            v = [t.__getattribute__(key) for t in trades]
            ser = pd.Series(data=v, index=None, dtype=type_map[key], name=key)
            ser_list[key] = ser
        df_trades = pd.DataFrame(ser_list)
        df_trades.index.name = 'index'
    
        trades_fn = os.path.join(folder_path, 'trades.csv')
        configs_fn = os.path.join(folder_path, 'configs.json')
        jutil.create_dir(trades_fn)
    
        df_trades.to_csv(trades_fn)
        jutil.save_json(self.props, configs_fn)
    
        print ("Backtest results has been successfully saved to:\n" + folder_path)
    
    '''
    def run_event(self):
        data_api = self.ctx.data_api
        universe = self.ctx.universe
        
        data_api.add_batch_subscribe(self, universe)
        
        self.ctx.trade_date = self.start_date
        
        def __extract(func):
            return lambda event: func(event.data, **event.kwargs)
        
        ee = self.ctx.strategy.eventEngine  # TODO event-driven way of lopping, is it proper?
        ee.register(EVENT.CALENDAR_NEW_TRADE_DATE, __extract(self.ctx.strategy.on_new_day))
        ee.register(EVENT.MD_QUOTE, __extract(self._process_quote_bar))
        ee.register(EVENT.MARKET_CLOSE, __extract(self.close_day))
        
        while self.ctx.trade_date <= self.end_date:  # each loop is a new trading day
            quotes = data_api.get_daily_quotes(self.ctx.trade_date)
            if quotes is not None:
                # gateway.oneNewDay()
                e_newday = Event(EVENT.CALENDAR_NEW_TRADE_DATE)
                e_newday.data = self.ctx.trade_date
                ee.put(e_newday)
                ee.process_once()  # this line should be done on another thread
                
                # self.ctx.strategy.onNewday(self.ctx.trade_date)
                self.ctx.strategy.pm.on_new_day(self.ctx.trade_date, self.last_date)
                self.ctx.strategy.ctx.trade_date = self.ctx.trade_date
                
                for quote in quotes:
                    # self.processQuote(quote)
                    e_quote = Event(EVENT.MD_QUOTE)
                    e_quote.data = quote
                    ee.put(e_quote)
                    ee.process_once()
                
                # self.ctx.strategy.onMarketClose()
                # self.closeDay(self.ctx.trade_date)
                e_close = Event(EVENT.MARKET_CLOSE)
                e_close.data = self.ctx.trade_date
                ee.put(e_close)
                ee.process_once()
                # self.ctx.strategy.onSettle()
                
            else:
                # no quotes because of holiday or other issues. We don't update last_date
                print "in trade.py: function run(): {} quotes is None, continue.".format(self.last_date)
            
            self.ctx.trade_date = self.go_next_trade_date(self.ctx.trade_date)
            
            # self.ctx.strategy.onTradingEnd()

    '''



# encoding: utf-8

"""
Modules relevant to data.
"""

from .dataapi import DataApi
from .dataservice import RemoteDataService, DataService
from .dataview import DataView
from .py_expression_eval import Parser


# we do not expose align and basic
__all__ = ['DataApi', 'DataService', 'RemoteDataService', 'DataView', 'Parser']